/**
 * @module DB
 */

/**
 * Callback of a query to the database.
 * @callback dbQueryCallback
 * @param {Object} err set when there was an error, null otherwise.
 * @param {Object} rows array of the selected rows in case of SELECT query, in
 * case of INSERT/UPDATE it contains a summary of the query (statusCode,
 * affetcedRows, ...)
 */

/**
 * Template function to execute a query with parameters.
 * @param {String} query the query to execute written in SQL
 * @param {String[]} params parameters for the query
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function queryTemplate(query, params, callback) {
  // gets a connection to the database from the connection pool
  pool.getConnection(function(err, connection) {
    // if the connection is not established it's impossible to proceed
    // with the query. call the callback, if defined, with the error
    if (err && callback) {
      callback(err, null);
    }

    connection.on('error', function(err) {
      if (callback) {
        callback(err, null);
      }
    });

    console.log("QUERY: " + query.replace(/\s+/g, ' '));
    console.log("PARAMS: ", params);

    // executes the query, closes the connection and calls the callback
    // if is defined
    connection.query(query, params, function(err, rows) {
      connection.release();
      if (callback) {
        callback(err, rows);
      }
    });
  });
}

/**
 * Creates a new user.
 * @param {Date} date creation date.
 * @param {String} username username of the user.
 * @param {String} staticPublicKey hex string of user's static public key.
 * @param {String} ephemeralPublicKey hex string of user's ephemeral public key.
 * @param {String} signingPublicKey hex string of user's signing public key.
 * @param {String} privateData Base64 of user's private data.
 * @param {String} salt salt used to generate the password-derived key.
 * @param {String} friendshipAddress address where friendship request should be send.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function createSessionUser(date, username, staticPublicKey, ephemeralPublicKey,
    signingPublicKey, privateData, salt, friendshipAddress, callback) {
  // TODO publicProfile now is forced to 0, which will be the default profile
  queryTemplate(
      "\
        INSERT INTO User (publicProfile, date, username, staticPublicKey, \
                          ephemeralPublicKey, signingPublicKey, privateData, \
                          salt, friendshipAddress) \
        VALUES(0, ?, ?, ?, ?, ?, ?, ?, ?) \
    ",
      [date, username, staticPublicKey, ephemeralPublicKey, signingPublicKey,
        privateData, salt, friendshipAddress], callback);
}

/**
 * Gets an user by username.
 * @param {String} username username of the user to search.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function getUserByUsername(username, callback) {
  queryTemplate(
      " \
        SELECT   id, date, username, staticPublicKey, ephemeralPublicKey, \
                 signingPublicKey, publicProfile, privateData, salt, \
                 friendshipAddress \
        FROM     User \
        WHERE    username = ? \
    ",
      [username], callback);
}

/**
 * Gets information of users.
 * @param {Integer[]} ids ids of the users to get.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function getUsers(ids, callback) {
  queryTemplate(
      "\
        SELECT   id, date, username, staticPublicKey, ephemeralPublicKey, \
                 signingPublicKey, publicProfile, friendshipAddress \
        FROM     User \
        WHERE    id IN (?) \
    ",
      [ids], callback);
}

/**
 * Gets information of users.
 * @param {Integer[]} ids ids of the users to get.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function getSessionUsers(ids, callback) {
  queryTemplate(
      "\
        SELECT   id, date, username, staticPublicKey, ephemeralPublicKey, \
                 signingPublicKey, publicProfile, privateData, salt, \
                 friendshipAddress \
        FROM     User \
        WHERE    id IN (?) \
    ",
      [ids], callback);
}

/**
 * Updates information of an user.
 * @param {Integer} id id of the user.
 * @param {Date} date modification date.
 * @param {Integer} publicProfile id of the user's public profile.
 * @param {String} ephemeralPublicKey Base64 of the user's ephemeral public key.
 * @param {String} privateData Base64 of user's private data.
 * @param {String} salt salt used to obtain the password-derived key.
 * @param {String} friendshipAddress address where friendship requests are sent.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function updateSessionUser(id, date, publicProfile, ephemeralPublicKey, privateData,
    salt, friendshipAddress, callback) {
  queryTemplate(
      "\
        UPDATE  User \
        SET     date = ?, publicProfile = ?, ephemeralPublicKey = ?, \
                privateData = ?, salt = ?, friendshipAddress = ? \
        WHERE   id = ? \
    ",
      [date, publicProfile, ephemeralPublicKey, privateData, salt,
       friendshipAddress, id], callback);
}

/**
 * Removes a version of an user.
 * @param {Integer} id the id of the user.
 * @param {String} date the version to remove.
 * @param {String} removeToken Base128 representation of the remove token.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function removeSessionUser(id, date, removeToken, callback) {
    queryTemplate(
        "\
          DELETE FROM User \
          WHERE       id = ? AND date = STR_TO_DATE(?, '%Y-%m-%dT%T.000Z') AND removeToken = ? \
     ",
       [id, date, removeToken], callback);
}

/**
 * Creates a new profile.
 * @param {Date} date creation date.
 * @param {String} privateData encrypted private data encoded in Base64
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function createProfile(date, privateData, callback) {
  queryTemplate(
      "\
        INSERT INTO Profile (date, privateData) \
        VALUES(?, ?) \
    ",
      [date, privateData], callback);
}

/**
 * Gets information of profiles.
 * @param {Integer[]} ids ids of the profiles to get.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function getProfiles(ids, callback) {
  queryTemplate(
      "\
        SELECT   id, date, privateData \
        FROM     Profile \
        WHERE    id IN (?) \
    ",
      [ids], callback);
}

/**
 * Updates information of a profile.
 * @param {Integer} id id of the profile.
 * @param {Date} date modification date.
 * @param {String} privateData Base64 of profile's private data.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function updateProfile(id, date, privateData, callback) {
  queryTemplate(
      "\
        UPDATE  Profile \
        SET     date = ?, privateData = ? \
        WHERE   id = ? \
    ",
      [date, privateData, id], callback);
}

/**
 * Removes a version of a profile.
 * @param {Integer} id the id of the user.
 * @param {Integer} date the version to remove.
 * @param {String} removeToken Base128 representation of the remove token.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function removeProfile(id, date, removeToken, callback) {
    queryTemplate(
        "\
          DELETE FROM Profile \
          WHERE       id = ? AND date = STR_TO_DATE(?, '%Y-%m-%dT%T.000Z') AND removeToken = ? \
     ",
       [id, date, removeToken], callback);
}

/**
 * Creates a new group.
 * @param {Date} date
 * @param {String} publicKey Base64 of group's public key.
 * @param {String} privateData Base64 of group's private data.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function createGroup(date, signingPublicKey, privateData, callback) {
  queryTemplate(
      "\
        INSERT INTO `Group` (date, signingPublicKey, privateData, adminData) \
        VALUES(?, ?, ?, '') \
    ",
      [date, signingPublicKey, privateData], callback);
}

/**
 * Gets information of groups.
 * @param {Integer[]} ids ids of the groups to get.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function getGroups(ids, callback) {
  queryTemplate(
      "\
        SELECT   id, date, signingPublicKey, privateData \
        FROM     `Group` \
        WHERE    id IN (?) \
    ",
      [ids], callback);
}

/**
 * Gets administrative information of groups.
 * @param {Integer[]} ids ids of the groups to get.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function getGroupsAdmin(ids, callback) {
  queryTemplate(
      "\
        SELECT   id, date, adminData AS privateData \
        FROM     `Group` \
        WHERE    id IN (?) \
    ",
      [ids], callback);
}

/**
 * Updates information of a group.
 * @param {Integer} id id of the group.
 * @param {Date} date modification date.
 * @param {String} privateData Base64 of group's private data.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function updateGroup(id, date, privateData, callback) {
  queryTemplate(
      "\
        UPDATE  `Group` \
        SET     date = ?, privateData = ? \
        WHERE   id = ? \
    ",
      [date, privateData, id], callback);
}

/**
 * Updates the administrative information of a group.
 * @param {Integer} id id of the group.
 * @param {Date} date modification date.
 * @param {String} adminData Base64 of group's private data.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function updateGroupAdmin(id, date, adminData, callback) {
  queryTemplate(
      "\
        UPDATE  `Group` \
        SET     date = ?, adminData = ? \
        WHERE   id = ? \
    ",
      [date, adminData, id], callback);
}

/**
 * Removes a version of a group.
 * @param {Integer} id the id of the user.
 * @param {Integer} date the version to remove.
 * @param {String} removeToken Base128 representation of the remove token.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function removeGroup(id, date, removeToken, callback) {
    queryTemplate(
        "\
          DELETE FROM Group \
          WHERE       id = ? AND date = STR_TO_DATE(?, '%Y-%m-%dT%T.000Z') AND removeToken = ? \
     ",
       [id, date, removeToken], callback);
}

/**
 * Creates a new message.
 * @param {Date} date modification date.
 * @param {String} address the address of the recipient.
 * @param {String} privateData Base64 of message's private data.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function createMessage(date, address, privateData,
    callback) {
  // FIXME: this is not safe, for better solutions see:
  //        https://stackoverflow.com/questions/5416548/mysql-two-column-primary-key-with-auto-increment
  //        Not fixing until we decide whether to switch to MyISAM or to a
  //        simpler and more controllable DBMS, which is better idea.
  queryTemplate(
      "\
        INSERT INTO Message (date, address, id, privateData) \
        VALUES(?, ?, (SELECT id FROM (SELECT COALESCE(MAX(id) + 1, 0) AS id FROM Message WHERE address = ?) AS Msg), ?); \
        SELECT MAX(id) AS id FROM Message WHERE address = ?; \
    ",
      [date, address, address, privateData, address], callback);
}

/**
 * Gets information of messages.
 * @param {Integer[]} ids ids of the messages to get.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function getMessages(ids, callback) {
  queryTemplate(
      "\
        SELECT   id, date, address, privateData \
        FROM     Message \
        WHERE    id IN (?) \
    ",
      [ids], callback);
}

/**
 * Retrieves all the messages created in a given range of time.
 * @param {String} address the address of the recipient.
 * @param {Date} dateFrom
 * @param {Date} dateTo
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
 // TODO: drop
function getMessagesByDate(address, dateFrom, dateTo,
    callback) {
  queryTemplate(
      "\
        SELECT  id, date, address, privateData \
        FROM    Message \
        WHERE   address = ? \
                date BETWEEN ? AND ? \
    ",
      [address, dateFrom, dateTo], callback);
}

/**
 * Retrieves the last messages received.
 * @param {String} address the address of the recipient for the messages.
 * @param {Integer} numMessages maximum number of elements to retrieve.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function getLastMessages(addresses, num, callback) {
  queryTemplate(
      "\
        SELECT  id, date, address, privateData \
        FROM    Message \
        WHERE   " + (["FALSE"].concat(addresses.map(function (obj) { return "(address = ?)" }))).join(" OR ") + " \
        ORDER BY id DESC \
        LIMIT   ? \
    ",
    addresses.concat([num]), callback);
}

/**
 * Retrieves the last messages received starting from a particular id.
 * @param {Array} addresses array of objects containing an address and id from
 *                which start to fetch.
 * @param {Integer} num maximum number of elements to retrieve.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function getMessagesAfterId(addresses, num, callback) {
  // TODO: create temporary table and join instead of a list of ORs?
  queryTemplate("\
        SELECT   id, date, address, privateData \
        FROM     Message \
        WHERE    " + (["FALSE"].concat(addresses.map(function (obj) { return "(address = ? AND id > ? )" }))).join(" OR ") + " \
        ORDER BY date DESC \
        LIMIT    ? \
    ",
    Array.prototype.concat.apply([], addresses.map(function (obj) { return [obj.address, obj.id]})).concat([num]), callback);
}

/**
 * Updates information of a message.
 * @param {Integer} id id of the message.
 * @param {Date} date
 * @param {String} privateData Base64 of group's private data.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
// TODO: allow to update the address field?
function updateMessage(id, date, privateData, callback) {
  queryTemplate(
      "\
        UPDATE  Message \
        SET     date = ?, privateData = ? \
        WHERE    id = ? \
    ",
      [date, privateData, id], callback);
}

/**
 * Removes a version of a message.
 * @param {Integer} id the id of the user.
 * @param {Integer} date the version to remove.
 * @param {String} removeToken Base128 representation of the remove token.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function removeMessage(id, date, removeToken, callback) {
    queryTemplate(
        "\
          DELETE FROM Message \
          WHERE       id = ? AND date = STR_TO_DATE(?, '%Y-%m-%dT%T.000Z') AND removeToken = ? \
     ",
       [id, date, removeToken], callback);
}

/**
 * Creates a new WoT entry.
 * @param {String} member HMAC of the friend id with the group key as key.
 * @param {String} group HMAC of the group id with the group key as key.
 * @param {String} editToken Base128 representation of the edit token.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function createWoTEntry(member, group, editToken, callback) {
  // You can insert a new record if the group does not exist
  // or if it exists and the editToken is the same.
  // DUAL is a mysql dummy table, used to make a conditional insert.
  // See http://stackoverflow.com/questions/913841/mysql-conditional-insert
  // Also: using backticks for `group` because reserved word
  queryTemplate(
      "\
        INSERT INTO WoT(member, `group`, editToken) \
        SELECT ?, ?, ? \
          FROM dual \
          WHERE NOT EXISTS ( \
            SELECT editToken, `group`, COUNT(*) AS c \
            FROM WoT WHERE `group` = ? \
            AND editToken <> ? \
            GROUP BY `group`, editToken \
            HAVING c > 0); \
    ",
      [member, group, editToken, group, editToken], callback);
}

/**
 * Gets the groups we share with a user.
 * @param {String[]} members HMACs of the request author's ID (one for every group)
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function getCommonGroups(members, callback) {
  queryTemplate(
      "\
        SELECT   member AS userIdMac, `group` AS groupIdMac \
        FROM     WoT \
        WHERE    member IN (?) \
    ",
      [members], callback);
}

/**
 * Removes an entire group from the WoT.
 * @param {String} editToken Base128 representation of the edit token.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function removeWoTEntries(editToken, callback) {
    queryTemplate(
        "\
          DELETE FROM WoT \
          WHERE       editToken = ? \
     ",
       [editToken], callback);
}

/**
 * Deletes all the data from all the tables. This is a debug method.
 * @param {dbQueryCallback} callback the callback that handles the query result.
 */
function cleanDb(callback) {
    var cleaner = function (table, callback) {
      return function () {
        queryTemplate(
            "\
              DELETE FROM " + table + "; \
         ", [], callback);
      };
    };
    cleaner("Group", cleaner("Message", cleaner("Profile", cleaner("User", callback))))();
}

var config = require('./config.js'),
    mysql = require('mysql');

var pool = mysql.createPool(config.mysqlConfig);

// XXX sql_modes needs to be 'TRADITIONAL' otherwise insert of NULL value
//     into NOT NULL fields does not fails (WTF?!?).
queryTemplate("SET SESSION sql_mode = 'TRADITIONAL'");

module.exports.createSessionUser = createSessionUser;
module.exports.getUserByUsername = getUserByUsername;
module.exports.getUsers = getUsers;
module.exports.getSessionUsers = getSessionUsers;
module.exports.updateSessionUser = updateSessionUser;
module.exports.removeSessionUser = removeSessionUser;

module.exports.createProfile = createProfile;
module.exports.getProfiles = getProfiles;
module.exports.updateProfile = updateProfile;
module.exports.removeProfile = removeProfile;

module.exports.createGroup = createGroup;
module.exports.getGroups = getGroups;
module.exports.getGroupsAdmin = getGroupsAdmin;
module.exports.updateGroup = updateGroup;
module.exports.updateGroupAdmin = updateGroupAdmin;
module.exports.removeGroup = removeGroup;

module.exports.createMessage = createMessage;
module.exports.getMessages = getMessages;
module.exports.getMessagesByDate = getMessagesByDate;
module.exports.getLastMessages = getLastMessages;
module.exports.getMessagesAfterId = getMessagesAfterId;
module.exports.updateMessage = updateMessage;
module.exports.removeMessage = removeMessage;

module.exports.createWoTEntry = createWoTEntry;
module.exports.getCommonGroups = getCommonGroups;
module.exports.removeWoTEntries = removeWoTEntries;

module.exports.cleanDb = cleanDb;
