#!/usr/bin/env node

var restify = require('restify'),
    socketio = require('socket.io'),
    sock = require('./socket'),
    config = require('./config');

var server = restify.createServer();
server.use(restify.acceptParser(server.acceptable));
server.use(restify.queryParser());
server.use(restify.bodyParser());
server.use(restify.gzipResponse());
server.use(restify.throttle(config.restifyThrottleConfig));
server.use(restify.CORS());
server.use(restify.fullResponse());

var io = socketio.listen(server);
config.socketIOConfig(io);

var parameters = {
  allowReset: false
};

process.argv.forEach(function (argument) {
  if (argument === "--allow-reset")
    parameters.allowReset = true;
});

io.on('connection', function(socket) {

  function routerError(errorMessage, res) {
    res({
      error: true,
      message: errorMessage
    });
  }

  function handleRequest(router, req, res) {
    if (!req) {
      routerError("Empty request", res);
    } else {
      var handler = router[req.object];

      if (!handler) {
        routerError("Method not available", res);
      } else {
        handler(req, res);
      }
    }
  }

  socket.on('create', function(req, res) {
    var router = {
      "session-user": sock.createSessionUser,
      "profile": sock.createProfile,
      "group": sock.createGroup,
      "message": sock.createMessage,
      "wot-entry": sock.createWoTEntry
    };
    handleRequest(router, req, res);
  });

  socket.on('read', function(req, res) {
    var router = {
      "username": sock.getUserByUsername,
      "session-user": sock.getSessionUsers,
      "user": sock.getUsers,
      "profile": sock.getProfiles,
      "group": sock.getGroups,
      "group-admin": sock.getGroupsAdmin,
      "message": sock.getMessages,
      "message-data": sock.getMessagesByDate,
      "last-messages": sock.getLastMessages,
      "messages-after-id": sock.getMessagesAfterId,
      "common-groups": sock.getCommonGroups
    };
    handleRequest(router, req, res);
  });

  socket.on('update', function(req, res) {
    var router = {
      "session-user": sock.updateSessionUser,
      "profile": sock.updateProfile,
      "group": sock.updateGroup,
      "group-admin": sock.updateGroupAdmin,
      "message": sock.updateMessage
    };
    handleRequest(router, req, res);
  });

  socket.on('delete', function(req, res) {
    var router = {
      "session-user": sock.removeSessionUser,
      "profile": sock.removeProfile,
      "group": sock.removeGroup,
      "message": sock.removeMessage,
      "wot-entries": sock.removeWoTEntries
    };
    handleRequest(router, req, res);
  });

  if (parameters.allowReset) {
    console.warn("Warning! The reset method is enable, use this only for " +
                 "debugging purposes!");
    socket.on('reset', function (req, res) {
      var router = {
        "clean": sock.cleanDb
      };
      handleRequest(router, req, res);
    });
  }
});

server.listen(config.serverListenPort, config.serverListenIP);
