/**
 * @module Socket
 */

/**
 * Callback of SocketIO request.
 * @callback socketIOCallback
 * @param {Object} resp response to send back to the client.
 */

/**
 * Generates an error message.
 * @param {Integer} code the code of the error.
 * @param {String} errorMessage the description of the error.
 * @returns {Object} the error object.
 */
function generateError(code, errorMessage) {
  return {
    error: true,
    code: code,
    message: errorMessage
  };
}

/**
 * Handles errors of a query to the database.
 * @param {Object} err the error object returned by a query to the database.
 * @param {Object} rows the result of a query to the database.
 * @param {Function} res the callback function to be called with a message to send.
 * @returns {Boolean} true if there was an error, false otherwise.
 */
function handleErrors(err, rows, res) {
  var restError = true;
  var errorCodes = {
    UnableHandleRequest: 0,
    TooFast: 1,
    DuplicateKey: 2,
    ParameterMissing: 3,
    ResourceNotPresent: 4
  };

  console.log("ERROR: ", err, "\n");

  if (err) {
    console.log(err);
    if (err.fatal) {
      res(generateError(errorCodes.UnableHandleRequest, "Unable to handle request"));
    } else {
      switch (err.code) {
        case "ER_DUP_ENTRY":
          res(generateError(errorCodes.DuplicateKey, "Duplicate key value"));
          break;
        case "ER_BAD_NULL_ERROR":
          res(generateError(errorCodes.ParameterMissing, "One or more parameters are missing"));
          break;
        default:
          res(generateError(errorCodes.UnableHandleRequest, "Unable to handle request"));
      }
    }
  } else {
    if (!rows) {
      res(generateError(errorCodes.ResourceNotPresent, "Resource not present"));
    } else if (rows.affectedRows === 0) {
      res(generateError(errorCodes.ResourceNotPresent, "Resource not present"));
    } else {
      restError = false;
    }
  }

  return restError;
}

/**
 * Gets the current date suppressing milliseconds.
 * @returns {Date} a Date object with zeroed milliseconds.
 */
function currentDate() {
  var result;

  result = new Date();

  // Truncate milliseconds, since MySQL ignores them
  result.setMilliseconds(0);

  return result;
}

/**
 * Creates a new user.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function createSessionUser(req, res) {
  var username = req.username,
      staticPublicKey = req.staticPublicKey,
      ephemeralPublicKey = req.ephemeralPublicKey,
      signingPublicKey = req.signingPublicKey,
      privateData = req.privateData,
      salt = req.salt,
      friendshipAddress = req.friendshipAddress,
      date = currentDate();

  if (!username.match(/^[a-zA-Z0-9]{3,}$/)) {
    res(generateError("Username not valid"));
    return;
  }

  db.createSessionUser(date, username, staticPublicKey, ephemeralPublicKey,
      signingPublicKey, privateData, salt,
      friendshipAddress, function(err, rows) {
    if (!handleErrors(err, rows, res)) {
      res({
        status: "Ok",
        id: rows.insertId,
        date: date
      });
    }
  });
}

/**
 * Gets an user by username.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function getUserByUsername(req, res) {
  var username = req.username;
  db.getUserByUsername(username, function(err, rows) {
    // Get user by username is the only query that sends an error if the is
    // no result
    var user = (rows && rows.length === 1) ? rows : undefined;
    if (!handleErrors(err, user, res)) {
      res(user);
    }
  });
}

/**
 * Gets information of users, excluding private fields.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function getUsers(req, res) {
  var ids = req.ids;
  db.getUsers(ids, function(err, rows) {
    if (!handleErrors(err, rows, res)) {
      res(rows);
    }
  });
}

/**
 * Gets information of users, including private fields.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function getSessionUsers(req, res) {
  var ids = req.ids;
  db.getSessionUsers(ids, function(err, rows) {
    if (!handleErrors(err, rows, res)) {
      res(rows);
    }
  });
}

/**
 * Updates information of an user.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function updateSessionUser(req, res) {
  var id = req.id,
      publicProfile = req.publicProfile,
      ephemeralPublicKey = req.ephemeralPublicKey,
      salt = req.salt,
      privateData = req.privateData,
      friendshipAddress = req.friendshipAddress,
      date = currentDate();
  db.updateSessionUser(id, date, publicProfile, ephemeralPublicKey, privateData,
      salt, friendshipAddress, function(err, rows) {
        if (!handleErrors(err, rows, res)) {
          res({
            status: "Ok",
            date: date
          });
        }
      });
}

/**
 * Removes a version of an user.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function removeSessionUser(req, res) {
  var id = req.id,
      date = req.date,
      removeToken = req.removeToken;
  db.removeSessionUser(id, date, removeToken, function (err, rows) {
      if (!handleErrors(err, rows, res)) {
        res({
          status: "Ok",
          date: date
        });
      }
  });
}

/**
 * Creates a new profile.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function createProfile(req, res) {
  var privateData = req.privateData,
      date = currentDate();
  db.createProfile(date, privateData, function(err, rows) {
    if (!handleErrors(err, rows, res)) {
      res({
        status: "Ok",
        id: rows.insertId,
        date: date
      });
    }
  });
}

/**
 * Gets information of profiles.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function getProfiles(req, res) {
  var ids = req.ids;
  db.getProfiles(ids, function(err, rows) {
    if (!handleErrors(err, rows, res)) {
      res(rows);
    }
  });
}

/**
 * Updates information of a profile.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function updateProfile(req, res) {
  var id = req.id,
      privateData = req.privateData,
      date = currentDate();
  db.updateProfile(id, date, privateData, function(err, rows) {
    if (!handleErrors(err, rows, res)) {
      res({
        status: "Ok",
        date: date
      });
    }
  });
}

/**
 * Removes a version of a profile.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function removeProfile(req, res) {
  var id = req.id,
      date = req.date,
      removeToken = req.removeToken;
  db.removeProfile(id, date, removeToken, function (err, rows) {
      if (!handleErrors(err, rows, res)) {
        res({
          status: "Ok",
          date: date
        });
      }
  });
}

/**
 * Creates a new group.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function createGroup(req, res) {
  var signingPublicKey = req.signingPublicKey,
      privateData = req.privateData,
      date = currentDate();
  db.createGroup(date, signingPublicKey, privateData, function(err, rows) {
    if (!handleErrors(err, rows, res)) {
      res({
        status: "Ok",
        id: rows.insertId,
        date: date
      });
    }
  });
}

/**
 * Gets information of groups.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function getGroups(req, res) {
  var ids = req.ids;
  db.getGroups(ids, function(err, rows) {
    if (!handleErrors(err, rows, res)) {
      res(rows);
    }
  });
}

/**
 * Gets administrative information of groups.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function getGroupsAdmin(req, res) {
  var ids = req.ids;
  db.getGroupsAdmin(ids, function(err, rows) {
    if (!handleErrors(err, rows, res)) {
      res(rows);
    }
  });
}

/**
 * Updates information of a group.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function updateGroup(req, res) {
  var id = req.id,
      privateData = req.privateData,
      date = currentDate();
  db.updateGroup(id, date, privateData, function(err, rows) {
    if (!handleErrors(err, rows, res)) {
      res({
        status: "Ok",
        date: date
      });
    }
  });
}

/**
 * Updates administrative information of a group.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function updateGroupAdmin(req, res) {
  var id = req.id,
      adminData = req.privateData,
      date = currentDate();
  db.updateGroupAdmin(id, date, adminData, function(err, rows) {
    if (!handleErrors(err, rows, res)) {
      res({
        status: "Ok",
        date: date
      });
    }
  });
}

/**
 * Removes a version of a group.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function removeGroup(req, res) {
  var id = req.id,
      date = req.date,
      removeToken = req.removeToken;
  db.removeGroup(id, date, removeToken, function (err, rows) {
      if (!handleErrors(err, rows, res)) {
        res({
          status: "Ok",
          date: date
        });
      }
  });
}

/**
 * Creates a new message.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function createMessage(req, res) {
  var address = req.address,
      privateData = req.privateData,
      date = currentDate();
  db.createMessage(date, address, privateData,
      function(err, rows) {
        if (!handleErrors(err, rows, res)) {
          res({
            status: "Ok",
            id: rows[1][0].id,
            date: date
          });
        }
      });
}

/**
 * Gets information of messages.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function getMessages(req, res) {
  var ids = req.ids;
  db.getMessages(ids, function(err, rows) {
    if (!handleErrors(err, rows, res)) {
      res(rows);
    }
  });
}

/**
 * Retrieves all the messages created in a given range of time.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function getMessagesByDate(req, res) {
  var address = req.address,
      from = new Date(req.from),
      to = new Date(req.to);
  db.getMessagesByDate(address, from, to,
      function(err, rows) {
        if (!handleErrors(err, rows, res)) {
          res(rows);
        }
      });
}

/**
 * Retrieves the last messages received.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function getLastMessages(req, res) {
  var addresses = req.addresses,
      num = req.num;
  db.getLastMessages(addresses, num, function(err, rows) {
    if (!handleErrors(err, rows, res)) {
      res(rows);
    }
  });
}

/**
 * Retrieves the last messages received starting from a particular id.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function getMessagesAfterId(req, res) {
  var addresses = req.addresses,
      num = Math.min(req.num, 100);
  db.getMessagesAfterId(addresses, num, function(err,
      rows) {
    if (!handleErrors(err, rows, res)) {
      res(rows);
    }
  });
}

/**
 * Updates information of a message.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function updateMessage(req, res) {
  var id = req.id,
      privateData = req.privateData,
      date = currentDate();
  db.updateMessage(id, date, privateData, function(err, rows) {
    if (!handleErrors(err, rows, res)) {
      res({
        status: "Ok",
        date: date
      });
    }
  });
}

/**
 * Removes a version of a message.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function removeMessage(req, res) {
  var id = req.id,
      removeToken = req.removeToken;
  db.removeMessage(id, removeToken, function (err, rows) {
      if (!handleErrors(err, rows, res)) {
        res({
          status: "Ok"
        });
      }
  });
}

/**
 * Creates a new WoT entry.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function createWoTEntry(req, res) {
  var member = req.userIdMac,
      group = req.groupIdMac,
      editToken = req.editToken;
  db.createWoTEntry(member, group, editToken, function(err, rows) {
    if (!handleErrors(err, rows, res)) {
      res({
        status: "Ok"
      });
    }
  });
}

/**
 * Gets the groups we share with a user.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function getCommonGroups(req, res) {
  var members = req.members;
  db.getCommonGroups(members, function(err, rows) {
    if (!handleErrors(err, rows, res)) {
      res(rows);
    }
  });
}

/**
 * Removes an entire group from the WoT.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function removeWoTEntries(req, res) {
  var editToken = req.editToken;
  db.removeWoTEntries(editToken, function (err, rows) {
      if (!handleErrors(err, rows, res)) {
        res({
          status: "Ok"
        });
      }
  });
}

/**
 * Completely cleans the database. This is a debug method.
 * @param {Object} req request parameters.
 * @param {socketIOCallback} res response callback.
 */
function cleanDb(req, res) {
  db.cleanDb(function (err, rows) {
    if (!handleErrors(err, rows, res)) {
      res({
        status: "Ok"
      });
    }
  });
}

var db = require('./db');

module.exports.createSessionUser = createSessionUser;
module.exports.getUserByUsername = getUserByUsername;
module.exports.getUsers = getUsers;
module.exports.getSessionUsers = getSessionUsers;
module.exports.updateSessionUser = updateSessionUser;
module.exports.removeSessionUser = removeSessionUser;

module.exports.createProfile = createProfile;
module.exports.getProfiles = getProfiles;
module.exports.updateProfile = updateProfile;
module.exports.removeProfile = removeProfile;

module.exports.createGroup = createGroup;
module.exports.getGroups = getGroups;
module.exports.getGroupsAdmin = getGroupsAdmin;
module.exports.updateGroup = updateGroup;
module.exports.updateGroupAdmin = updateGroupAdmin;
module.exports.removeGroup = removeGroup;

module.exports.createMessage = createMessage;
module.exports.getMessages = getMessages;
module.exports.getMessagesByDate = getMessagesByDate;
module.exports.getLastMessages = getLastMessages;
module.exports.getMessagesAfterId = getMessagesAfterId;
module.exports.updateMessage = updateMessage;
module.exports.removeMessage = removeMessage;

module.exports.createWoTEntry = createWoTEntry;
module.exports.getCommonGroups = getCommonGroups;
module.exports.removeWoTEntries = removeWoTEntries;

module.exports.cleanDb = cleanDb;
